"""
Classes for describing asset transforms to backends.
"""

__all__ = ['BaseTransform']


class TransformMeta(type):
    """
    Meta class for `BaseTransform` to that an `_id` has been specified for each
    transform class and
    """

    def __new__(meta, name, bases, dct):

        cls = super(TransformMeta, meta).__new__(meta, name, bases, dct)

        # Register the class so we can look it up when restoring from JSON
        if name is not 'BaseTransform':
            assert '_id' in dct, 'No `_id` class attribute set for transform'
            BaseTransform._register[dct['_id']] = cls

        return cls


class BaseTransform(metaclass=TransformMeta):
    """
    Transform classes provide a mechanism for describe how to transform an
    asset, for example resizing, rotating, cropping, ect.

    Each backend asset manager is responsible for translating these Transform
    into the necessary instuctions for the storage service it represents.

    A transform is made up of a unique Id and the settings that will be used
    when applying it. Settings should be set as a table and values should be
    JSON safe.

    The transform's Id is set as a class property, normally in the following
    format:

        `{asset_type}.{transform}` -> `image.crop`

    service specific transforms should be suffixed, e.g:

        `{asset_type}.{transform}.{service}` -> `image.face.hangar51`

    """

    # Base transforms have no Id and should not be used directly as a transform
    _id = ''

    # A table of all registered transform classes by Id
    _register = {}

    def __init__(self, settings=None):

        # The settings that will be used when applying the transform
        if settings is None:
            settings = {}

        self._settings = settings

    @property
    def id(self):
        # Return the transforms Id
        return self._id

    @property
    def settings(self):
        # Return the settings to use when applying the transform
        return self._settings.copy()

    # Public methods

    def to_json_type(self):
        """
        Return a dictionary for the transform with values converted to JSON safe
        types.
        """
        return {'id': self.id, 'settings': self.settings}

    # Class methods

    @staticmethod
    def from_json_type(data):
        """
        Convert a dictionary of JSON safe type values to a transform instance.
        """

        # Find the associated transform class
        transform_cls = BaseTransform._register[data['id']]

        # Initialize a new instance of the class using the settings stored in
        # data.
        return transform_cls(**data['settings'])


"""
Generic gallery chain.

: `gallery_field`
    The field against the document that stores the gallery of assets (required).

: `projection`
    The projection used when requesting the document from the database (defaults
    to None which means the detault projection for the frame class will be
    used).

    NOTE: The `gallery_field` must be one of the projected fields otherwise the
    gallery will always appear to be empty.

: `validators`
    A list of validators (see manhattan.assets.validators) that will be used to
    validate assets within to the gallery (defaults to None, as in no
    validation).

"""

import json

import flask
from manhattan.assets import Asset
from manhattan.assets.fields import AssetField
from manhattan.assets.transforms.base import BaseTransform
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, validators
from manhattan.nav import Nav, NavItem
from werkzeug.datastructures import MultiDict

from manhattan.manage.views import (
    factories as manage_factories,
    utils as manage_utils
)

__all__ = ['gallery_chains']


# Define the chains
gallery_chains = ChainMgr()

# GET
gallery_chains['get'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'get_assets',
    'decorate',
    'render_template'
    ])

# POST
gallery_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'get_assets',
    'validate',
    [
        [
            'store_assets',
            'redirect'
        ],
        [
            'decorate',
            'render_template'
        ]
    ]
])


# Define the links
gallery_chains.set_link(manage_factories.config(
    gallery_field=None,
    gallery_render_kw=None,
    gallery_validators=None
    ))
gallery_chains.set_link(manage_factories.authenticate())
gallery_chains.set_link(manage_factories.get_document())
gallery_chains.set_link(manage_factories.render_template('gallery.html'))
gallery_chains.set_link(manage_factories.redirect('view', include_id=True))

@gallery_chains.link
def decorate(state):
    """
    Add decor information to the state (see `utils.base_decor` for further
    details on what information the `decor` dictionary consists of).

    This link adds a `decor` key to the state.
    """
    document = state[state.manage_config.var_name]
    state.decor = manage_utils.base_decor(
        state.manage_config,
        state.view_type,
        document
    )

    # Title
    state.decor['title'] = state.manage_config.titleize(document)

    # Breadcrumbs
    if Nav.exists(state.manage_config.get_endpoint('list')):
        state.decor['breadcrumbs'].add(
            manage_utils.create_breadcrumb(state.manage_config, 'list')
        )
    if Nav.exists(state.manage_config.get_endpoint('view')):
        state.decor['breadcrumbs'].add(
            manage_utils.create_breadcrumb(
                state.manage_config,
                'view',
                document
            )
        )
    state.decor['breadcrumbs'].add(NavItem('Gallery'))

@gallery_chains.link
def get_assets(state):
    """
    Get the asset information for the gallery from the document (GET) or the
    request (POST).

    This link adds `assets` to the state which contains the list of assets to
    be stored against the gallery field, and `assets_json_type` which is a used
    in the template to provide the frontend JS with a serialzied version of the
    assets.
    """
    document = state[state.manage_config.var_name]

    # Get the existing assets
    assets = []
    assets_table = {}
    for asset in (getattr(document, state.gallery_field) or []):
        if not isinstance(asset, Asset):
            asset = Asset(asset)
        assets.append(asset)
        assets_table[asset.key] = asset

    # Get the list of updated assets
    state.base_transforms_modified = {}

    if flask.request.method == 'POST':

        # Merge existing with the updated assets
        assets = []
        updated_assets = json.loads(flask.request.form.get('assets'))

        for asset in updated_assets:

            preview_uri = None
            if 'preview_uri' in asset:
                preview_uri = asset.pop('preview_uri')

            if not isinstance(asset, Asset):
                asset = Asset(asset)

            base_transforms = asset.base_transforms
            user_meta = asset.user_meta

            if asset.key in assets_table:
                existing_asset = assets_table[asset.key]

                # Set any user defined meta data against the asset
                existing_asset.user_meta = user_meta

                # For the base transforms we perform a comparison to flag that
                # they have been modified and therefore we need to regenerate
                # existing ssets.
                old = json.dumps(existing_asset.base_transforms)
                new = json.dumps(asset.base_transforms)
                state.base_transforms_modified[existing_asset.key] = old != new

                asset = existing_asset

            else:
                # Find the temporary asset
                asset = flask.current_app.asset_mgr.get_temporary_by_key(
                    asset.key
                )

            # Set any user defined meta data against the asset
            asset.user_meta = user_meta
            asset.base_transforms = base_transforms

            if asset:
                assets.append(asset)

            # Cater for a preview URI provided client-side (required to provide
            # a preview after the field validates but the form fails to).
            if preview_uri:
                asset.preview_uri = preview_uri

    state.assets = assets
    state.assets_json_type = [a.to_json_type() for a in state.assets]

@gallery_chains.link
def validate(state):
    """
    Validate the gallery of assets.

    If there's an error against one or more of the assets in the gallery then
    this link will add `asset_errors` to the state. This is dictionary if
    errors with the asset `key` property as the key and the error message as
    the value.
    """

    # Check at least one validators has been specified
    if not state.gallery_validators:
        return True

    # Define a form against which we can perform the validation
    class AssetForm(BaseForm):

        asset = AssetField(
            'Asset',
            validators=[validators.Required()] + state.gallery_validators
        )

    # Validate every asset in the gallery
    valid = True
    for asset in state.assets:
        form = AssetForm(
            MultiDict({'asset': json.dumps(asset.to_json_type())}),
            asset=None if asset.temporary else asset
        )
        if not form.validate():
            flask.flash(
                asset.filename + ': ' + ','.join(form.errors['asset']),
                'error'
                )
            valid = False

    return valid

@gallery_chains.link
def store_assets(state):
    """
    Convert temporary assets to permenant assets and store any other changes
    to asset information.
    """
    asset_mgr = flask.current_app.asset_mgr
    document = state[state.manage_config.var_name]

    assert state.gallery_field, 'No gallery field defined'

    # Generate variations for new and updated assets
    for asset in state.assets:

        if not (asset.temporary or \
                state.base_transforms_modified.get(asset.key)):
            continue

        if asset.temporary:
            # Store the asset permenantly (converting it from a temporary
            # asset).

            # FIXME: This shouldn't be redefined here (just asset_mgr)
            flask.current_app.asset_mgr.store(asset)

        variations = {}

        # Check for variations defined against the manage config
        if hasattr(state.manage_config, state.gallery_field + '_variations'):
            variations = getattr(
                state.manage_config,
                state.gallery_field + '_variations'
            )

        # Check for existing variations (which overide any variation set
        # against the manage config.
        if asset.variations:
            for name, variation_asset in asset.variations.items():

                if not isinstance(variation_asset, Asset):
                    variation_asset = Asset(variation_asset)

                variations[name] = [BaseTransform.from_json_type(t)
                    for t in variation_asset.local_transforms]

        # Ensure the draft variation is never updated
        if '--draft--' in variations:
            variations.pop('--draft--')

        if variations:

            # Get the list of base transforms for the asset
            base_transforms = [BaseTransform.from_json_type(t)
                    for t in asset.base_transforms]

            # Store variations for the asset
            asset_mgr.generate_variations(
                asset,
                variations,
                base_transforms
            )

    # Update the database
    if hasattr(document, 'logged_update'):
        assets = [a.to_json_type() for a in state.assets]
        document.logged_update(
            state.manage_user,
            {state.gallery_field: assets}
        )

    else:
        setattr(document, state.gallery_field, state.assets)
        document.update(state.gallery_field)

    flask.flash('{document} updated.'.format(document=document))



"""
Classes for storing assets.
"""

import mimetypes
import os
from urllib.parse import urljoin, urlparse

import flask
from mongoframes import SubFrame
from mongoframes.factory.makers.images import ImageURL

__all__ = ['Asset']


# Additional common mimetypes not part of the standard mimetypes library
mimetypes.add_type('text/csv', '.csv')
mimetypes.add_type('image/webp', '.webp')


class Asset(SubFrame):
    """
    The `Asset` class provides a representation of assets as embedded documents
    with the database. The asset data is stored separately via an asset backend,
    but information about each asset including their URLs are stored in the
    database and access via the asset class.
    """

    # The root location for assets, this value is used to generate the assets
    # `path` and `url` (URL) properties.
    _asset_root = ''

    _fields = {
        # A flag indicating if this is the base asset or (if False) a variation
        # of a base asset.
        'base',

        # A unique key (at least within the scope of the application) for the
        # asset.
        'key',

        # The assets filename
        'filename',

        # The type of asset, must be either 'file' or 'image'
        'type',

        # A table of meta data describing the asset (provided by the backend
        # service).
        'core_meta',

        # A table of user defined meta information for the asset
        'user_meta',

        # A flag indicating if the asset is temporary. Temporary assets are
        # typically created prior to an asset being saved permenantly. They
        # allow the asset's information to be held against a single unique key
        # which can easily be transferred between client and server.
        #
        # On committing a change temporary assets are converted to 'permenant'
        # assets and any variations for the asset are generated.
        #
        # NOTE: The `temporary` flag is only set for base assets, variations
        # don't require the flag as there existance is based on their parent
        # base asset.
        'temporary',

        # A stack of base transforms applied to all varations (but not the base
        # asset). This allows user manipulations of the base asset to be applied
        # to any variation.
        'base_transforms',

        # A stack of local transforms applied to the asset (only applicable to
        # variations). This allows variations to be regenerated.
        'local_transforms',

        # A table of variations for the asset. A variation is also an asset,
        # based on the base variation but transformed (e.g resized, rotated,
        # etc.) The `variations` property holds a named set of assets each of
        # which is a varation of the (base) asset.
        'variations'
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # By default assign an empty dictionary to the `core_meta`, `user_meta`
        # and `variations` fields if not specified.
        if not self.core_meta:
            self.core_meta = {}

        if not self.user_meta:
            self.user_meta = {}

        if not self.base_transforms:
            self.base_transforms = []

        if not self.local_transforms:
            self.local_transforms = []

        if not self.variations:
            self.variations = {}

        # Set up access to the assets meta data
        self._meta = AssetMeta(self)

        # A preview URI for the asset set when an asset has been modified
        # client-side. This value is never stored in the database and is used
        # to allow a preview to persist between multiple form submissions.
        self.preview_uri = None

    def __str__(self):
        return self.url

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.key == other.key
        return False

    @property
    def content_type(self):
        # Return a content type for the asset based on the extension
        return mimetypes.guess_type(self.filename)[0]

    @property
    def ext(self):
        # Return the assets file extension
        return os.path.splitext(self.filename)[1]

    @property
    def fake_url(self, size=None):
        # Return a fake URL for the asset

        try:
            size = None
            # Check if the assets size is given
            if 'image' in self.meta and 'size' in self.meta.image:
                size = self.meta.image['size']
            else:
                # If no size is available default to something small
                size = [500, 500]

            # Check for fake URL configuration settings
            options = flask.current_app.config.get('ASSET_FAKE_MAKER_OPTIONS', {})
        except Exception as e:
            print(e)

        return ImageURL(size[0], size[1], **options)._assemble()

    @property
    def meta(self):
        # Return a special Meta class that will manage core and user meta for
        # the asset.
        return self._meta

    @meta.setter
    def meta(self, value):
        for k, v in value:
            self._meta[k] = v

    @property
    def path(self):
        # Return the relative path for the asset
        return urljoin(urlparse(self._asset_root).path, self.filename)

    @property
    def url(self):
        # Return the absolute URL for the asset

        # If the current flask application is configured to be aware of fake
        # assets then we check for fake assets and return a dummy URL.
        if flask.current_app.config.get('ASSET_ENABLE_FAKES'):
            if self.core_meta.get('__manhattan_fake__'):
                return self.fake_url

        return urljoin(self._asset_root, self.filename)

    @property
    def v(self):
        # Short-cut property for variations. Can be used anywhere but was added
        # primarily to reduce clutter in templates.
        return self.variations

    def to_json_type(self):
        document_dict = super().to_json_type()

        # Add the asset's content type
        document_dict['content_type'] = self.content_type

        # Add the asset's extension
        document_dict['ext'] = self.ext

        # Add the asset's URL
        document_dict['url'] = self.url

        # Add the preview URI for the asset if present
        if self.preview_uri:
            document_dict['preview_uri'] = self.preview_uri

        return document_dict

    @classmethod
    def default_projection(cls):
        """
        Short-cut for defining a simple projection, usage:

            projection = {
                'image': Asset.default_projection()
            }

        """
        projection = {'$sub': cls}
        projection.update(cls._default_projection)
        return projection


# Whilst the default projection isn't typically used directly it is useful when
# defining projection for classes with assets fields. For example:
#
#    _default_projection = {
#        'image': {'$sub': assets.Asset, **assets.Asset._default_projection}
#        }
#
# See the `default_projection` shortcut class method.
#
Asset._default_projection = {'variations': {'$sub.': Asset}}


class AssetMeta:
    """
    The `AssetMeta` class provides a mechanism for managing asset meta data
    using both the core and user defined meta data.

    By default values are returned from the user defined meta data but if not
    present then the core meta data value is used.

    Meta values are always set against the user defined meta data.
    """

    def __init__(self, asset):

        # The asset meta is being managed for
        self._asset = asset

    def __getattr__(self, name):
        if '_asset' in self.__dict__:
            asset = self.__dict__['_asset']

            # Check user defined meta data
            if name in asset.user_meta:
                return asset.user_meta[name]

            # Check core meta data
            if name in asset.core_meta:
                return asset.core_meta[name]

        return None

    def __setattr__(self, name, value):
        if '_asset' in self.__dict__:
            asset = self.__dict__['_asset']
            if value is None:
                # Remove value if set to `None`
                if name in asset.user_meta:
                    del asset.user_meta[name]
            else:
                # Set the value
                asset.user_meta[name] = value
        else:
            super(AssetMeta, self).__setattr__(name, value)

    def __getitem__(self, name):
        asset = self.__dict__['_asset']

        # Check user defined meta data
        if name in asset.user_meta:
            return asset.user_meta[name]

        # Check core meta data
        return asset.core_meta.get(name, None)

    def __contains__(self, name):
        asset = self.__dict__['_asset']
        return name in asset.user_meta or name in asset.core_meta

    def get(self, name, default=None):
        asset = self.__dict__['_asset']

        # Check user defined meta data
        if name in asset.user_meta:
            return asset.user_meta[name]

        # Check core meta data
        return asset.core_meta.get(name, None)



"""
A collection of image transforms.
"""

from manhattan.assets.transforms.base import BaseTransform

__all__ = [
    'Crop',
    'Face',
    'Fit',
    'Output',
    'Rotate',
    'SmartCrop'
    ]


class Crop(BaseTransform):
    """
    Apply a crop to an image.

    Each value (top, left, bottom, right) should be between 0.0 and 1.0
    representing a normalized value for the given dimension.
    """

    _id = 'image.crop'

    def __init__(self, top, left, bottom, right):
        super().__init__({
            'top': top,
            'left': left,
            'bottom': bottom,
            'right': right
            })


class Face(BaseTransform):
    """
    Detect and crop a face from within the image.
    """

    _id = 'image.face'

    def __init__(self, bias=None, padding=0, min_padding=0):
        super().__init__({
            'bias': bias,
            'padding': padding,
            'min_padding': min_padding
            })


class Fit(BaseTransform):
    """
    Resize an image to fit within a set of dimensions. The width and height
    should be specified in pixels.
    """

    _id = 'image.fit'

    def __init__(self, width, height=None):
        super().__init__({'width': width, 'height': height or width})


class Output(BaseTransform):
    """
    Set the output format of an image. The list of support formats will vary
    depending on service, however as a minimum the following formats should be
    supported by all services:

    - jpg (supports quality)
    - gif
    - png
    - webp (supports quality)
    """

    _id = 'image.output'

    def __init__(self, format, quality=None):
        settings = {'format': format}
        if quality is not None:
            settings['quality'] = quality

        super().__init__(settings)


class Rotate(BaseTransform):
    """
    Rotate an image by 90, 180 or 270 degrees.
    """

    _id = 'image.rotate'

    def __init__(self, angle):
        super().__init__({'angle': angle})


class SmartCrop(BaseTransform):
    """
    Cropping a region of the image matching the given aspect ratio (w / h)
    centered around a detected area of interest (based on the detection of a
    face or feature).
    """

    _id = 'image.smart_crop'

    def __init__(self, aspect_ratio):
        super().__init__({'aspect_ratio': aspect_ratio})


"""
A set of maker classes for producing fake asset data.
"""

import random

import flask
from mongoframes.factory.makers import Faker, Maker
from shortuuid import ShortUUID

from manhattan.assets import Asset

__all__ = [
    'FileAsset',
    'ImageAsset'
    ]


class FileAsset(Maker):
    """
    Generate a fake file asset.
    """

    def __init__(self, extensions):
        # The file extension to use for the generate file asset
        self._extensions = extensions

    def _assemble(self):
        key = FileAsset.generate_uid(6)
        filename = Faker.get_fake().slug()
        return {
            'base': True,
            'key': key,
            'filename': '{filename}.{key}.{extension}'.format(
                extension=random.choice(self._extensions),
                filename=filename,
                key=key
                ),
            'type': 'file',
            'core_meta': {
                '__manhattan_fake__': True,
                'filename': filename,
                'length': int(abs(random.gauss(5000000, 3000000)))
                },
            'user_meta': {},
            'temporary': False,
            'variations': {}
            }

    def _finish(self, value):
        return Asset(value)

    @staticmethod
    def generate_uid(length):
        """Generate a uid of a given length"""
        su = ShortUUID(alphabet='abcdefghijklmnopqrstuvwxyz0123456789')
        return su.uuid()[:length]


class ImageAsset(Maker):
    """
    Generate a fake image asset.
    """

    def __init__(self, size, variations=None):
        # The size of the original image in the form of a 2 element list, e.g:
        # `[width, height]`.
        self._size = size

        # The variations that should exist for the Asset. Commonly these are
        # passed directly from the manage config class, e.g:
        #
        #     image = ImageAsset(
        #         [1000, 1000],
        #         variations=ProductCfg.image_variations)
        #         )
        #
        # However any valid dictionary of named variations can be provided, e.g:
        #
        #     image = ImageAsset(
        #         [1000, 1000],
        #         variations={'listing': [Fit(200, 200), Output('jpg', 90)]}
        #         )
        #
        self._variations = variations or {}

    def _assemble(self):
        # Generate a base image asset
        key = FileAsset.generate_uid(6)
        extension = 'jpg'
        filename = Faker.get_fake().slug()
        size = [int(self._size[0]), int(self._size[1])]
        asset = {
            'base': True,
            'key': key,
            'filename': '{filename}.{key}.{extension}'.format(
                extension=extension,
                filename=filename,
                key=key
                ),
            'type': 'image',
            'core_meta': {
                '__manhattan_fake__': True,
                'filename': filename,
                'image': {
                    'size': size,
                    'mode': 'RGB'
                    },
                'length': size[0] * size[1] * 3
                },
            'user_meta': {},
            'temporary': False,
            'transforms': [],
            'variations': {}
            }

        # Generate variations of the base asset
        filename_str = '{filename}.{key}.{name}.{version}.{extension}'

        # Build the list of required variations
        for name, transforms in self._variations.items():

            # Determine the size and output format of the variation
            variation_extension = extension
            variation_size = list(size)
            version = FileAsset.generate_uid(3)

            # We're only concerned with changes to the size of the image and the
            # output format.
            for transform in transforms:
                # Crop
                if transform.id == 'image.crop':
                    crop = transform.settings
                    variation_size = [
                        int(variation_size[0] * (crop['right'] - crop['left'])),
                        int(variation_size[1] * (crop['bottom'] - crop['top']))
                        ]

                # Fit
                elif transform.id == 'image.fit':
                    fit = transform.settings
                    scale = min(
                        fit['width'] / variation_size[0],
                        fit['height']/ variation_size[1]
                        )
                    variation_size[0] = int(scale * variation_size[0])
                    variation_size[1] = int(scale * variation_size[1])

                # Output
                elif transform.id == 'image.output':
                    extension = transform.settings['format']

                # Rotate
                elif transform.id == 'image.rotate':
                    if transform.setttings.angle in [90, 270]:
                        variation_size = [variation_size[1], variation_size[0]]

            # Create the variation
            asset['variations'][name] = {
                'base': False,
                'key': '.'.join([key, version]),
                'filename': filename_str.format(
                    extension=extension,
                    filename=filename,
                    key=key,
                    name=name,
                    version=version
                    ),
                'type': 'image',
                'core_meta': {
                    '__manhattan_fake__': True,
                    'filename': filename,
                    'image': {
                        'size': variation_size,
                        'mode': 'RGB'
                        },
                    'length': variation_size[0] * variation_size[1] * 3
                    },
                'user_meta': {},
                'temporary': False,
                'variations': {}
                }

        return asset

    def _finish(self, value):
        asset = Asset(value)
        for name, variation in asset.variations.items():
            asset.variations[name] = Asset(variation)
        return asset



import json

import flask
from manhattan.forms import fields
from wtforms import widgets

__all__ = (
    'AssetField'
    )


class AssetField(fields.Field):
    """
    A field that supports a file being uploaded.
    """

    widget = widgets.HiddenInput()

    def __init__(self, label=None, validators=None, render_kw=None, **kwargs):

        # Ensure the field is flagged as an asset field
        if not render_kw:
            render_kw = {}

        render_kw['data-mh-file-field'] = True

        # Flag indicating if the asset's base transforms have changed (which
        # would indicate that the asset's variations need to be regenerated).
        self.base_transform_modified = False

        super().__init__(label, validators, render_kw=render_kw, **kwargs)

    def _value(self):
        if self.raw_data:
            return ' '.join(self.raw_data)
        elif self.data is not None:
            return json.dumps(self.data.to_json_type())
        return ''

    def process_data(self, value):
        """Process the the fields initial or default value"""
        self.data = value

    def process_formdata(self, values):
        """Process a value(s) submitted to the form"""
        if not values or not values[0]:
            self.data = None
            return

        # Convert the serialized asset JSON to an asset
        asset_data = json.loads(values[0])
        key = asset_data['key']
        base_transforms = asset_data.get('base_transforms')
        user_meta = asset_data.get('user_meta')

        # Check if the asset is a new temporary asset or is the asset initial set
        # for the field.
        if self.data and self.data.key == key:
            asset = self.data

        # If the asset is a temporary asset then attempt to retrieve it's details
        # from the temporary cache.
        else:
            asset = flask.current_app.asset_mgr.get_temporary_by_key(key)

        if not asset:
            self.data = None
            return

        # Set any user defined meta data against the asset
        asset.user_meta = user_meta

        # For the base transforms we perform a comparison to flag that they
        # have been modified and therefore we need to regenerate existing
        # assets.
        old = json.dumps(asset.base_transforms)
        new = json.dumps(base_transforms)
        self.base_transform_modified = old != new

        asset.base_transforms = base_transforms

        # Cater for a preview URI provided client-side (required to provide
        # a preview after the field validates but the form fails to).
        if asset_data.get('preview_uri'):
            asset.preview_uri = asset_data['preview_uri']

        self.data = asset



from wtforms.validators import ValidationError

__all__ = [
    'AssetType',
    'ContentType',
    'ImageSize',
    ]


class AssetType(object):
    """
    Validate that an asset is a given asset type (either file or image).
    """

    def __init__(self, asset_type, message=None):

        # The type the asset must be
        self.asset_type = asset_type

        # The error message to display if the asset isn't the required type
        self.message = message

    def __call__(self, form, field):
        # Check if there's anything to validate
        if not field.data:
            return

        # Check if the asset type is the required type
        if field.data.type == self.asset_type:
            return

        # Raise a validation error flagging the assets's type is not the require
        # type.
        message = self.message
        if message is None:
            message = field.gettext('The asset must be of type: {asset_type}.')

        raise ValidationError(message.format(asset_type=self.asset_type))


class ContentType(object):
    """
    Validate that an asset is belongs to a set of given content(mime) types.
    """

    def __init__(self, content_types, message=None):

        # The set of content types the asset's content type must belong to
        self.content_types = set(content_types)

        # The error message to display if the asset isn't the required type
        self.message = message

    def __call__(self, form, field):
        # Check if there's anything to validate
        if not field.data:
            return

        # Check if the asset's content type belongs to the required set
        if field.data.content_type in self.content_types:
            return

        # Raise a validation error flagging the assets's content type does not
        # belong to the required set.
        message = self.message
        if message is None:
            message = field.gettext(
                "The asset's content type must be one of: {content_types}")
        content_types = ', '.join(sorted(list(self.content_types)))
        raise ValidationError(message.format(content_types=content_types))


class ImageSize(object):
    """
    Validate that an image asset's size is within a min/max range. If either end
    of the range is not given then the following defaults are assumed:

    - min: `[1, 1]`
    - max: `[99999, 99999]`

    NOTE: The min/max sizes should be given as a 2 element list of the form
    ``[width, height]`
    """

    def __init__(self, min_size=None, max_size=None, message=None):
        # The minimum size the image asset must be
        self.min_size = min_size
        if self.min_size is None:
            self.min_size = [1, 1]

        # The maximum size the image asset must be
        self.max_size = max_size
        if self.max_size is None:
            self.max_size = [99999, 99999]

        # The error message to display if the asset isn't the required type
        self.message = message

    def __call__(self, form, field):
        # Check if there's anything to validate
        if not field.data:
            return

        # Check the asset is an image
        asset = field.data
        if asset.type != 'image':
            return

        # Check the asset provides information on the size of the image
        size = None
        if asset.meta.image and 'size' in asset.meta.image:
            size = asset.meta.image['size']
        if not size:
            return

        # Check the image is within the required range (min/max)
        too_small = False
        too_big = False
        if size[0] < self.min_size[0] or size[1] < self.min_size[1]:
            too_small = True
        elif size[0] > self.max_size[0] or size[1] > self.max_size[1]:
            too_big = True

        if not (too_small or too_big):
            return

        # Raise a validation error flagging the image assets's size is not
        # within the required range.
        message = self.message
        size = self.min_size if too_small else self.max_size
        if message is None:
            if too_small:
                message = field.gettext(
                    "Image is too small, must be at least: {size}")
            else:
                message = field.gettext(
                    "Image is too big, must be less than: {size}")

        raise ValidationError(message.format(size=size))


from .assets import *
from . import fields
from . import makers
from . import static
from . import transforms
from . import validators


from manhattan.assets import Asset
from manhattan.assets.views.transform import transform_chains
from manhattan.assets.views.upload import upload_chains


__all__ = ['Assets']


class Assets:
    """
    The `Assets` class provides the initialization code for the package.
    """

    def __init__(
        self,
        app,
        root,
        backend,
        settings=None,
        upload_path='/upload-asset',
        transform_path='/transform-asset'
        ):
        self._app = app

        # Configure the URL root for assets
        Asset._asset_root = root

        # Set up the asset manager
        self._app.asset_mgr = backend.AssetMgr(**(settings or {}))

        # Set up views for managing assets

        # Upload
        self._app.add_url_rule(
            upload_path,
            endpoint='upload_asset',
            view_func=upload_chains.copy().flask_view(),
            methods=['POST']
        )

        # Transform
        self._app.add_url_rule(
            transform_path,
            endpoint='transform_asset',
            view_func=transform_chains.copy().flask_view(),
            methods=['POST']
        )



"""
Asset backend for Hangar51.
"""

import os

from box51 import Box51, Box51Exception

from manhattan.assets import Asset
from manhattan.assets.backends.base import BaseAssetMgr
from manhattan.assets.backends.exceptions import RetrieveError, StoreError
from manhattan.assets.transforms.base import BaseTransform

__all__ = ['AssetMgr']


class AssetMgr(BaseAssetMgr):
    """
    Asset manager using the Box51 library.
    """

    _transform_converters = {
        'image.crop': lambda t: [
            'crop',
            [
                t.settings['top'],
                t.settings['right'],
                t.settings['bottom'],
                t.settings['left']
            ]
        ],
        'image.fit': lambda t: [
            'fit',
            [
                t.settings['width'],
                t.settings['height']
            ]
        ],
        'image.output': lambda t: ['output', t.settings],
        'image.rotate': lambda t: ['rotate', t.settings['angle']]
    }

    def __init__(self, asset_root, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._client = Box51(asset_root)

    def generate_variations(self, asset, variations, base_transforms=None):
        """Generate variations for the asset"""

        base_transforms = base_transforms or []

        # Generate the variations
        box51_variations = {}
        for name, transforms in variations.items():
            variation = []

            variation = self.convert_transforms(transforms)
            if base_transforms:
                variation = self.convert_transforms(base_transforms) \
                    + variation

            box51_variations[name] = variation

        try:
            result = self._client.generate_variations(
                asset.key,
                box51_variations
            )
        except Box51Exception as e:
            raise StoreError(str(e))

        # Store the variations against the asset
        prefix_path = ''
        if asset.temporary:
            prefix_path = self._client.TMP_DIR

        for name, variation in result.items():
            local_transforms = variations[name]
            asset_variation = Asset(
                base=False,
                key=variation['store_key'],
                filename=os.path.join(prefix_path, variation['store_key']),
                type=asset.type,
                core_meta=variation['meta'],
                local_transforms=[
                    BaseTransform.to_json_type(t) for t in local_transforms]
                )
            asset.variations[name] = asset_variation

        asset.base_transforms = [
            BaseTransform.to_json_type(t) for t in base_transforms]

        # If the asset is a temporary asset then update the asset cache
        self.update_cache(asset)

    def remove(self, asset):
        """Remove the specified asset"""
        try:
            self._client.remove(asset.key)
        except Box51Exception as e:
            raise StoreError(str(e))

    def retrieve(self, asset):
        """Retrieve the asset (the file)"""

        try:
            data = self._client.retrieve(asset.key)
        except Box51Exception as e:
            raise RetrieveError(str(e))
        return data

    def store_temporary(self, file, name=None):
        """Store an asset temporarily"""

        try:
            result = self._client.store(file, name=name, temporary=True)
        except Box51Exception as e:
            raise StoreError(str(e))

        # Create an asset representing the file
        tmp_path = self._client.TMP_DIR
        asset = Asset(
            base=True,
            key=result['store_key'],
            filename=os.path.join(tmp_path, result['store_key']),
            type=result['type'],
            core_meta=result['meta'],
            temporary=True
            )

        # Store the asset as a temporary asset
        self.update_cache(asset)

        return asset

    def store(self, file_or_asset, name=None):
        """
        Store an asset.

        NOTE: The `name` argument is ignored if an asset is provided, to rename
        an existing asset you must clone the asset with a new name and then
        store the resulting temporary asset.
        """
        asset = None
        if isinstance(file_or_asset, Asset):
            asset = file_or_asset
            asset.temporary = False

            # Remove the assets expiry date
            try:
                filename_remap = self._client.make_permanent(asset.key)
            except Box51Exception as e:
                raise StoreError(str(e))

            # Clear any reference to the temporary asset
            self.clear_cache(asset)

            # Switch asset and varation to use the new permenant paths
            asset.filename = filename_remap[asset.filename]
            for variation in asset.variations.values():
                variation.filename = filename_remap[variation.filename]

        else:

            # Store the file
            try:
                result = self._client.store(file_or_asset, name=name)
            except Box51Exception as e:
                raise StoreError(str(e))

            # Create an asset representing the file
            asset = Asset(
                base=True,
                key=result['store_key'],
                filename=result['store_key'],
                type=result['type'],
                core_meta=result['meta']
            )

        return asset



"""
Exceptions that should be raised when failing to store or retrieve an asset via
a backend service.
"""

__all__ = [
    # Exceptions
    'RetrieveError',
    'StoreError'
    ]


class RetrieveError(Exception):
    """
    Raised when a backend services fails to retreve an asset.
    """


class StoreError(Exception):
    """
    Raised when a backend services fails to store an asset.
    """


"""
Classes for implementing a backend for asset management.
"""

import json

from manhattan.assets import Asset

__all__ = ['BaseAssetMgr']


# Classes

class BaseAssetMgr:
    """
    Assets are managed via an asset manager class which must provide the API
    defined by the `BaseAssetMgr` class.
    """

    # A table of functions that convert transforms to a format understood by the
    # backend service.
    _transform_converters = None

    def __init__(self, cache, expires=3600):
        # The cache used to store references to temporary assets, must support
        # `get` and `set` methods, `set` must support an `timeout` argument
        # when the data will expire.
        #
        # `werkzeug.contrib.cache` provides a number of suitable classes for
        # creating caches (e.g `SimpleCache` and `MemcachedCache`).
        self._cache = cache

        # The amount of time before temporary assets expire
        self._expires = expires

    def clear_cache(self, asset):
        """Clear the given asset from the cache"""
        self._cache.delete(asset.key)

    def clone(self, asset, name=None):
        """Returning a temporary asset cloned from the specified asset"""

        # Retrieve the asset we're cloning
        file = self.retrieve(asset)

        # Store the file as a new temporary asset (cloning it)
        cloned = self.store_temporary((file, asset.filename), name=name)

        return cloned

    def generate_variations(self, asset, variations, base_transforms=None):
        """Generate variations for the asset"""
        raise NotImplementedError()

    def get_temporary_by_key(self, key):
        """Get a temporary asset by it's key"""
        asset_json = self._cache.get(key)
        if asset_json:
            asset = Asset(json.loads(asset_json))
            for name, variation in asset.variations.items():
                asset.variations[name] = Asset(variation)
            return asset

    def remove(self, asset):
        """Remove the specified asset"""
        raise NotImplementedError()

    def retrieve(self, asset):
        """Retrieve the asset"""
        raise NotImplementedError()

    def store_temporary(self, file, name=None):
        """Store an asset temporarily"""
        raise NotImplementedError()

    def store(self, file_or_asset, name=None):
        """
        Store an asset.

        NOTE: The `name` argument is ignored if an asset is provided, to rename
        an existing asset you must clone the asset with a new name and then
        store the resulting temporary asset.
        """
        raise NotImplementedError()

    def update_cache(self, asset):
        """Update the given asset in the cache"""
        if asset.temporary:
            asset_json = json.dumps(asset.to_json_type())
            self._cache.delete(asset.key)
            self._cache.set(asset.key, asset_json, self._expires)

    # Class methods

    @classmethod
    def convert_transforms(cls, transforms):
        """
        Convert a stack of transforms to a format understood by the backend
        service.
        """
        return [cls._transform_converters[t.id](t) for t in transforms]

    @classmethod
    def define_transform_converter(cls, transform_id, func):
        """Define a converter for the given transform (id)"""
        if cls._transform_converters is None:
            cls._transform_converters = {}
        cls._transform_converters[transform_id] = func



from collections import namedtuple

from .gallery import gallery_chains
from .transform import transform_chains
from .upload import upload_chains

__all__ = ['generic']


# We name space generic views using a named tuple to provide a slightly nicer
# way to access them, e.g:
#
#     from manhattan.manage.views import generic
#
#     view = generic.add
#
# And to make it easy to iterate through the list of generic views to make
# changes, e.g:
#
#     def authenticate(state):
#         """A custom authenticator for my site"""
#
#         ...
#
#     for view in generic:
#         view.set_link(authenticate)

# Define the named tuple (preventing the list of generic views being altered)
Generic = namedtuple(
    'Generic',
    [
        'gallery',
        'transform',
        'upload'
    ])

# Create an instance of Generic containing all the generic views
generic = Generic(
    gallery=gallery_chains,
    transform=transform_chains,
    upload=upload_chains
    )



from flask import current_app
from .assets import Asset

__all__ = [
    'min_file_projection',
    'min_image_projection'
]


def min_file_projection(field, extra=None):
    """Return a projection that will select minimal file info"""
    projection = {}
    projection[field] = {
        '$sub': Asset,
        'key': True,
        **(extra or {})
    }
    projection['.'.join([field, 'filename'])] = True
    projection['.'.join([field, 'core_meta', 'filename'])] = True
    projection['.'.join([field, 'core_meta', 'length'])] = True

    return projection

def min_image_projection(field, variations, extra=None):
    """Return a projection that will only select minimal image info"""
    projection = {}
    projection[field] = {
        '$sub': Asset,
        'variations': {'$sub.': Asset},
        **(extra or {})
    }
    projection['.'.join([field, 'filename'])] = True
    projection['.'.join([field, 'core_meta', 'image'])] = True
    projection['.'.join([field, 'core_meta', 'image', 'size'])] = True
    projection['.'.join([field, 'user_meta'])] = True

    for variation in variations:
        variation_path = '.'.join([field, 'variations', variation])
        projection[variation_path] = True
        projection['.'.join([variation_path, 'filename'])] = True

        core_meta_path = '.'.join([variation_path, 'core_meta'])
        projection[core_meta_path] = True
        projection['.'.join([core_meta_path, 'image'])] = True
        projection['.'.join([core_meta_path, 'image', 'size'])] = True
        projection['.'.join([core_meta_path, '__manhattan_fake__'])] = True

        user_meta_path = '.'.join([variation_path, 'user_meta'])
        projection[user_meta_path] = True

    return projection


"""
Asset backend for Hangar51.
"""

from datetime import datetime, timedelta
import json

from hangar51client import Hangar51Client, Hangar51ClientException, Variation

from manhattan.assets import Asset
from manhattan.assets.backends.base import BaseAssetMgr
from manhattan.assets.backends.exceptions import RetrieveError, StoreError
from manhattan.assets.transforms.base import BaseTransform

__all__ = ['AssetMgr']


class AssetMgr(BaseAssetMgr):
    """
    Asset manager using the Hangar51 service API.
    """

    # A table of functions that convert transforms to a format understood by the
    # backend service.
    _transform_converters = {
        'image.crop': lambda t: [
            'crop',
            [
                t.settings['top'],
                t.settings['right'],
                t.settings['bottom'],
                t.settings['left']
            ]
        ],
        'image.face': lambda t: ['face', t.settings],
        'image.fit': lambda t: [
            'fit',
            [
                t.settings['width'],
                t.settings['height']
            ]
        ],
        'image.output': lambda t: ['output', t.settings],
        'image.rotate': lambda t: ['rotate', t.settings['angle']],
        'image.smart_crop': lambda t: [
            'smart_crop',
            t.settings['aspect_ratio']
        ]
    }

    def __init__(self, api_key, api_endpoint=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set up an client for the Hangar51 API
        if api_endpoint:
            self._client = Hangar51Client(api_key, api_endpoint=api_endpoint)
        else:
            self._client = Hangar51Client(api_key)

    def generate_variations(self, asset, variations, base_transforms=None):
        """Generate variations for the asset"""

        base_transforms = base_transforms or []

        # Convert the variations
        h51_variations = {}
        for name, transforms in variations.items():
            variation = Variation()

            variation.ops = self.convert_transforms(transforms)
            if base_transforms:
                variation.ops = self.convert_transforms(base_transforms) \
                     + variation.ops

            h51_variations[name] = variation

        # Generate the variations
        try:
            result = self._client.generate_variations(asset.key, h51_variations)
        except Hangar51ClientException as e:
            raise StoreError(str(e))

        # Store the variations against the asset
        for name, variation in result.items():
            local_transforms = variations[name]
            asset_variation = Asset(
                base=False,
                key='{0}.{1}'.format(asset.key, variation['version']),
                filename=variation['store_key'],
                type=asset.type,
                core_meta=variation['meta'],
                local_transforms=[
                    BaseTransform.to_json_type(t) for t in local_transforms]
                )
            asset.variations[name] = asset_variation

        asset.base_transforms = [
            BaseTransform.to_json_type(t) for t in base_transforms]

        # If the asset is a temporary asset then update the asset cache
        self.update_cache(asset)

    def remove(self, asset):
        """Remove the specified asset"""
        try:
            self._client.set_expires(asset.key, expires=datetime.now())
        except Hangar51ClientException as e:
            raise StoreError(str(e))

    def retrieve(self, asset):
        """Retrieve the asset (the file)"""
        try:
            data = self._client.download(asset.key)
        except Hangar51ClientException as e:
            raise RetrieveError(str(e))
        return data

    def store_temporary(self, file, name=None):
        """Store an asset temporarily"""

        # Store the file
        expires = datetime.now() + timedelta(0, self._expires)
        try:
            result = self._client.upload(file, name=name, expires=expires)
        except Hangar51ClientException as e:
            raise StoreError(str(e))

        # Create an asset representing the file
        asset = Asset(
            base=True,
            key=result['uid'],
            filename=result['store_key'],
            type=result['type'],
            core_meta=result['meta'],
            temporary=True
            )

        # Store the asset as a temporary asset
        self.update_cache(asset)

        return asset

    def store(self, file_or_asset, name=None):
        """
        Store an asset.

        NOTE: The `name` argument is ignored if an asset is provided, to rename
        an existing asset you must clone the asset with a new name and then
        store the resulting temporary asset.
        """

        asset = None
        if isinstance(file_or_asset, Asset):
            asset = file_or_asset
            asset.temporary = False

            # Remove the assets expiry date
            try:
                self._client.set_expires(asset.key)
            except Hangar51ClientException as e:
                raise StoreError(str(e))

            # Clear any reference to the temporary asset
            self.clear_cache(asset)

        else:
            # Store the file
            try:
                result = self._client.upload(file_or_asset, name=name)
            except Hangar51ClientException as e:
                raise StoreError(str(e))

            # Create an asset representing the file
            asset = Asset(
                base=True,
                key=result['uid'],
                filename=result['store_key'],
                type=result['type'],
                core_meta=result['meta']
                )

        return asset



"""
Generic upload asset chain.

NOTE: The generic upload chain is designed to handle a single file upload,
multiple file uploads client side should result in multiple calls to the upload
chain.
"""

import flask
from manhattan.assets.transforms.images import Fit, Output
from manhattan.assets.backends import exceptions
from manhattan.chains import Chain, ChainMgr

from manhattan.manage.views import factories
from manhattan.manage.views import utils

__all__ = ['upload_chains']


# Define the chains
upload_chains = ChainMgr()

# POST
upload_chains['post'] = Chain([
    'config',
    'authenticate',
    'store_temporary_asset',
    'render_json'
])


# Define the links
upload_chains.set_link(factories.config())
upload_chains.set_link(factories.authenticate())

@upload_chains.link
def store_temporary_asset(state):
    """
    Store the file uploaded as a temporary asset.

    The client request is expected to include a file under the parameter name of
    `file`.

    If the file is successfully stored this links adds an `asset` key to the
    state containing the `Asset` instance representing the uploaded file.

    IMPORTANT: Validation of the asset itself is the responsibility of the view
    that is storing the associated parent document, not the upload chain. The
    upload chain merely validates that a file was provided and could be stored.
    """

    # Check a file was provided
    if 'file' not in flask.request.files \
            or not flask.request.files['file'].filename:
        return utils.json_fail('No file sent')

    # Attempt to store the file
    file = flask.request.files['file']
    asset_mgr = flask.current_app.asset_mgr
    try:
        state.asset = asset_mgr.store_temporary((file.filename, file))
    except exceptions.StoreError as e:
        return utils.json_fail(str(e))

    # Check to see if an image file type has been requested and if so validate
    # the asset backend recognized it as an image.
    file_type = flask.request.values.get('file_type')
    if file_type == 'image':
        if state.asset.type != file_type:
            return utils.json_fail('Not a supported: ' + file_type)

    # If the asset is an image then create a base variation
    if state.asset.type == 'image':

        # The `--base--` variation is used to allow the user to perform and
        # preview transforms against the image, a `Fit` transform is applied to
        # the image initially to restrict the size of the image and help with
        # performance.
        #
        # The `--thumb--` variation is used to provide a thumbnail view of the
        # image.
        try:
            config = flask.current_app.config
            draft_size = config.get('ASSET_DRAFT_SIZE', [1200, 1200])
            thumb_size = config.get('ASSET_THUMB_SIZE', [480, 480])

            draft_transforms = [
                Fit(draft_size[0], draft_size[1]),
                Output('jpg', 50)
                ]
            thumb_transforms = [
                Fit(thumb_size[0], thumb_size[1]),
                Output('jpg', 75)
                ]

            variations = {'--draft--': draft_transforms}
            if not flask.request.values.get('no-thumb'):
                variations['--thumb--'] = thumb_transforms

            asset_mgr.generate_variations(state.asset, variations)

            # Store the transform information against the variations
            draft_transforms = [t.to_json_type() for t in draft_transforms]
            state.asset.variations['--draft--'].local_transforms = \
                    draft_transforms

            thumb_transforms = [
                t.to_json_type() for t in thumb_transforms]
            state.asset.variations['--thumb--'].local_transforms = \
                    thumb_transforms

            asset_mgr.update_cache(state.asset)

        except exceptions.StoreError as e:
            return utils.json_fail(str(e))

@upload_chains.link
def render_json(state):
    """
    Return a successful response with the uploaded `asset` included in the
    payload.
    """
    return utils.json_success({'asset': state.asset.to_json_type()})


"""
Generic transform asset chain.
"""

import json

import flask
from manhattan.assets import Asset
from manhattan.assets.transforms import *
from manhattan.assets.transforms.base import BaseTransform
from manhattan.assets.backends import exceptions
from manhattan.chains import Chain, ChainMgr

from manhattan.manage.views import factories
from manhattan.manage.views import utils

__all__ = ['transform_chains']


# Define the chains
transform_chains = ChainMgr()

# POST
transform_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_asset',
    'transform_asset',
    'render_json'
])


# Define the links
transform_chains.set_link(factories.config())
transform_chains.set_link(factories.authenticate())

@transform_chains.link
def get_asset(state):
    """
    Get a temporary asset from the cache by key.

    This link adds the `asset` key to the state which contains the temporary
    asset looked up.
    """
    asset_mgr = flask.current_app.asset_mgr

    # Attempt to find the temporary asset by key
    key = flask.request.values.get('key', '')
    asset = asset_mgr.get_temporary_by_key(key)
    if not asset:
        return utils.json_fail('Asset not found')

    # Store the asset against the state
    state.asset = asset

@transform_chains.link
def transform_asset(state):
    """
    Set the base transforms for a temporary asset and update all associated
    variations.
    """
    asset_mgr = flask.current_app.asset_mgr

    # Convert the transform param from JSON to JSON type data
    transform_str = flask.request.values.get('transforms', '[]')
    try:
        json_transforms = json.loads(transform_str)
    except ValueError:
        return utils.json_fail('`transforms` is not valid JSON')

    # Convert the transforms from JSON type data to asset transforms
    base_transforms = [
        BaseTransform.from_json_type(t) for t in json_transforms]

    # Set the base transforms to the asset
    state.asset.base_transforms = json_transforms

    # Build a map of variations for each child variation
    local_transforms = {}
    full_transforms = {}
    for name, variation in state.asset.variations.items():

        # Extract the existing transforms
        local_transforms[name] = variation.local_transforms

        # Build the full transform required for the variation
        full_transforms[name] = base_transforms
        full_transforms[name] += [
            BaseTransform.from_json_type(t) for t in local_transforms[name]]

    # Regenerate the variations for the asset
    asset_mgr.generate_variations(state.asset, full_transforms)

    # Add the transform information back to each new variation
    for name in local_transforms.keys():
        state.asset.variations[name].local_transforms = local_transforms[name]

    # Update the asset in the cache
    asset_mgr.update_cache(state.asset)

@transform_chains.link
def render_json(state):
    """
    Return a successful response with the transformed `asset` included in the
    payload.
    """
    return utils.json_success({'asset': state.asset.to_json_type()})


"""
Functions for handling static assets.

The handling of static assets is likely to be project specific, however we
provide support for the default approach we use at getme on the basis that it's
simple to understand and deploy no matter you static asset pipeline.
"""

import json
import os


def get_static_asset(filepath):
    """
    Return a function for handling static assets.

    The returned function will look for a JSON file using the provided
    `filepath`. The JSON file (what we refer to as an asset pack) will contain
    a map of file references and their physical location, e.g:

        {'site.js': 'http://assets.mysite.co.uk/site.s3c8f6.js'}

    When the returned function is called (typically within a Jinja template) it
    will be passed a file reference and return the phsyical location.
    """

    # The date/time the asset pack was last modified
    last_modified = 0

    # A map of file references to physical locations
    file_map = {}

    # Only attempt to load the asset map on start up
    if os.path.exists(filepath):
        with open(filepath) as f:
            file_map = json.load(f)

    def _get_static_asset(file_ref):

        # Check if the asset is being served by a live server
        if '__live__' in file_map:
            return file_map['__live__'] + file_ref

        return file_map.get(file_ref, file_ref)

    return _get_static_asset